#include "gfx.h"
#include "globals.h"
#include "sim.h"

#include "x-fun.h"
#include "x-gfx.h"
#include "x-loadlum.h"
#include "x-vector3d.h"

#include "words.h"
#include "pane.h"

#include <iostream>
#include <vector>
using namespace std;

//-----------------------------------------------------------------------------
// function prototypes
//-----------------------------------------------------------------------------
void idleFunc();
void displayFunc();
void reshapeFunc( int width, int height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );
void activeMouseFunc( int x, int y);
void passiveMouseFunc(int x, int y);
void specialFunc( int key, int x, int y );

void initialize_graphics();
void initialize_simulation();
bool initialize_data();

void renderBackground();
void blendPane();
void updateNodeEntities();
void renderNodeEntities();

void savePastFreqs(string fileName);
void loadPastFreqs(string fileName);

//-----------------------------------------------------------------------------
// name: gfx_init( )
// desc: graphics/interaction entry point
//-----------------------------------------------------------------------------
bool gfx_init( int argc, const char ** argv )
{
#ifdef __APPLE__
    // save working dir
    char * cwd = getcwd( NULL, 0 );
    // set path
#else
    // store this globally
    Globals::path = cwd;
#endif
    
    // initialize GLUT
    glutInit( &argc, (char **)argv );
    
#ifdef __APPLE__
    //restore working dir
    chdir( cwd );
    free( cwd );
#endif
    
    // print about
    about();
    endline();
    
    // double buffer, use rgb color, enable depth buffer
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    // initialize the window size
    glutInitWindowSize( Globals::windowWidth, Globals::windowHeight );
    // set the window postion
    glutInitWindowPosition( 100, 100 );
    // create the window
    glutCreateWindow( "Scattered");
    // full screen
    // if( Globals::fullscreen ) glutFullScreen();
    
    // set the idle function - called when idle
    glutIdleFunc( idleFunc );
    // set the display function - called when redrawing
    glutDisplayFunc( displayFunc );
    // set the reshape function - called when client area changes
    glutReshapeFunc( reshapeFunc );
    // set the keyboard function - called on keyboard events
    glutKeyboardFunc( keyboardFunc );
    // set the mouse function - called on mouse stuff
    glutMouseFunc( mouseFunc );
    // set mouse function
    glutMotionFunc( activeMouseFunc );
    // set mouse function - called in passive motion
    glutPassiveMotionFunc( passiveMouseFunc );
    // for arrow keys, etc
	glutSpecialFunc (specialFunc );
    
    // do our own initialization
    initialize_graphics();
    // simulation
    initialize_simulation();
    // do data
    if( !initialize_data() )
    {
        // done
        return false;
    }
    
    return true;
}


//-----------------------------------------------------------------------------
// name: gfx_loop( )
// desc: hand off to graphics loop
//-----------------------------------------------------------------------------
void gfx_loop()
{
    // let GLUT handle the current thread from here
    glutMainLoop();
}


//-----------------------------------------------------------------------------
// Name: initialize_graphics( )
// Desc: sets initial OpenGL states and initializes any application data
//-----------------------------------------------------------------------------
void initialize_graphics()
{
    // log
    cerr << "Initializing graphics system..." << endl;
    
    // reset time
    XGfx::resetCurrentTime();
    // set simulation speed
    XGfx::setDeltaFactor( 1.0f );
    // get the first
    XGfx::getCurrentTime( true );
    // random
    XFun::srand();
    
    // set the GL clear color - use when the color buffer is cleared
    glClearColor( Globals::bgColor.actual().x, Globals::bgColor.actual().y, Globals::bgColor.actual().z, 1.0f );
    // set the shading model to 'smooth'
    glShadeModel( GL_SMOOTH );
    // enable depth
    glEnable( GL_DEPTH_TEST );
    // set the front faces of polygons
    glFrontFace( GL_CCW );
    // set fill mode
    glPolygonMode( GL_FRONT_AND_BACK, Globals::fillmode );
    // enable lighting
    glEnable( GL_LIGHTING );
    // enable lighting for front
    glLightModeli( GL_FRONT_AND_BACK, GL_TRUE );
    // material have diffuse and ambient lighting
    glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
    // enable color
    glEnable( GL_COLOR_MATERIAL );
    // normalize (for scaling)
    glEnable( GL_NORMALIZE );
    
    // enable light 0
    glEnable( GL_LIGHT0 );
    
    // setup and enable light 1
    glLightfv( GL_LIGHT1, GL_AMBIENT, Globals::light1_ambient );
    glLightfv( GL_LIGHT1, GL_DIFFUSE, Globals::light1_diffuse );
    glLightfv( GL_LIGHT1, GL_SPECULAR, Globals::light1_specular );
    glEnable( GL_LIGHT1 );
    
    // clear the color buffer once
    glClear( GL_COLOR_BUFFER_BIT );
}


//-----------------------------------------------------------------------------
// name: initialize_simulation( )
// desc: simulation setup
//-----------------------------------------------------------------------------
void initialize_simulation()
{
    // instantiate simulation
    Globals::sim = new Sim();

    // load words
    Globals::wordList = new Words();
    Globals::wordList->initialize(Globals::wordFile);
    
    cerr << "adding " << Globals::wordList->getNWords() << " words ..." << endl;
    // add words to simulation
    for (auto& a: Globals::wordList->existingWordsMap) Globals::sim->root().addChild(a.second->yText);
    
    // if applicable, first word is focal word
    if (Globals::wordList->existingWordsMap.size() > 0){
        Globals::focalWord = Globals::wordList->existingWordsMap.begin()->second;
        Globals::focalWord->initializeConfidence();
        
        Vector3D pos;
        Globals::focalWord->getPosFinal(& pos);
        Globals::viewX.update(pos[0],VIEW_SLEWRATE);
        Globals::viewY.update(pos[1],VIEW_SLEWRATE);
        Globals::viewZ.update(pos[2],VIEW_SLEWRATE);
        Globals::viewEyeX.y = 0;
        Globals::viewEyeY.y = 0;
    }
    
    // load control panel
    Globals::controlPanel = new Pane();
    Globals::controlPanel->initialize();
    if (Globals::focalWord) {
        Globals::controlPanel->setCirclePos(Globals::focalWord->getConfidence());
        string word = Globals::focalWord->getString();
        for (int i = 0; i < word.length(); i++) Globals::controlPanel->write(MODE_ENTERWORD, word[i]);
    }
    Globals::sim->root2D().addChild( Globals::controlPanel );
    
}

//-----------------------------------------------------------------------------
// name: initialize_data( )
// desc: load initial data
//-----------------------------------------------------------------------------
bool initialize_data()
{
    // randomize seed
    srand((unsigned int) time(NULL));
    
    // initialize synthesizer
    wsSynth * synth;
    for (int i = 0; i < NNOTES; i++){
        synth = new wsSynth();
        synth->initialize();
        Globals::synth[i] = synth;
        Globals::synthNote[i] = 0;
        if (Globals::focalWord){
            Globals::focalWord->setSynth(Globals::synth.at(i));
            Globals::focalWord->initializeConfidence();
        }
    }

    Globals::synthInd = 0;
    
    // initialize PCA values
    Globals::wordList->getPCASettings(Globals::pcaFile);
    
    // initialize past frequency, tempo data
    Globals::eighthNoteTempo = Globals::focalWord->getTempo();
    for (int i = 0; i < NPASTFREQS; i++){
        Globals::pastFreqs[i] = Globals::focalWord->getFreqs(i);
    }
    Globals::wordList->resetRandFreqs();
    
    // done
    return true;
}


//-----------------------------------------------------------------------------
// name: about()
// desc: ...
//-----------------------------------------------------------------------------
void about()
{
    line();
    fprintf( stderr, "wordSynth\n" );
    line();
    fprintf( stderr, "   | by Wendy Nie\n" );
    fprintf( stderr, "   | version: %s\n", Globals::version.c_str() );
}




//-----------------------------------------------------------------------------
// name: keys()
// desc: ...
//-----------------------------------------------------------------------------
void keys()
{
    line();
    fprintf( stderr, "run-time control\n" );
    line();
    fprintf( stderr, "  'h' - print this help message\n" );
    fprintf( stderr, "  's' - toggle fullscreen\n" );
    fprintf( stderr, "  'b' - toggle animated trails\n" );
    fprintf( stderr, "  'f' - toggle fog rendering\n" );
    fprintf( stderr, "  '[' and ']' - rotate automaton\n" );
    fprintf( stderr, "  '-' and '+' - zoom away/closer to center of automaton\n" );
    fprintf( stderr, "  'n' and 'm' - adjust amount of blending\n" );
    fprintf( stderr, "  '<' and '>' - adjust fog density\n" );
    fprintf( stderr, "  ''' - bg: white\n" );
    fprintf( stderr, "  ';' - bg: black\n" );
    fprintf( stderr, "  ',' - bg: blue\n" );
    fprintf( stderr, "  '.' - bg: gray\n" );
    fprintf( stderr, "  'q' - quit\n" );
}




//-----------------------------------------------------------------------------
// name: help()
// desc: ...
//-----------------------------------------------------------------------------
void help()
{
    endline();
    keys();
    endline();
}




//-----------------------------------------------------------------------------
// name: usage()
// desc: ...
//-----------------------------------------------------------------------------
void usage()
{
    line();
    fprintf( stderr, "command line arguments\n" );
    line();
    fprintf( stderr, "usage: wordSynth --[options] [name]\n" );
    fprintf( stderr, "   [options] = help | fullscreen" );
}




//-----------------------------------------------------------------------------
// name: endline()
// desc: ...
//-----------------------------------------------------------------------------
void endline()
{
    fprintf( stderr, "\n" );
}




//-----------------------------------------------------------------------------
// name: line()
// desc: ...
//-----------------------------------------------------------------------------
void line()
{
    fprintf( stderr, "---------------------------------------------------------\n" );
}




//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( int w, int h )
{
    // save the new window size
    Globals::windowWidth = w; Globals::windowHeight = h;
    // map the view port to the client area
    glViewport( 0, 0, w, h );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
}




//-----------------------------------------------------------------------------
// Name: look( )
// Desc: ...
//-----------------------------------------------------------------------------
void look( int view )
{
    // go
    Globals::fov.interp( XGfx::delta() );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    
    // create the viewing frustum
    switch (view) {
        case W_PERS :
            gluPerspective( Globals::fov.value, (GLfloat)Globals::windowWidth / (GLfloat)Globals::windowHeight, .005, 500.0 );
            break;
        case W_ORTHO :
            glOrtho(-Globals::windowWidth/(double)Globals::windowHeight, Globals::windowWidth/(double)Globals::windowHeight, -1, 1, 1, -1 );
            break;
        case W_2D :
            glOrtho(0.0f, Globals::windowWidth, Globals::windowHeight, 0.0f, 0.0f, 1.0f);
            break;
    }
    
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity();
    // position the view point
    float x = sin( Globals::viewEyeX.x );
    float y = sin( Globals::viewEyeY.x );
    float z = cos( Globals::viewEyeX.x ) * cos( Globals::viewEyeY.x );
    float mult = ( cos( Globals::viewEyeY.x ) < 0 ? -1.0f : 1.0f );
    
    gluLookAt( Globals::viewX.value + Globals::viewRadius.x * x ,
              Globals::viewY.value + Globals::viewRadius.x * y ,
              Globals::viewZ.value + Globals::viewRadius.x * z,
              Globals::viewX.value, Globals::viewY.value, Globals::viewZ.value,
              -x ,
              mult,
              -z );
    
    // set the position of the lights
    glLightfv( GL_LIGHT0, GL_POSITION, Globals::light0_pos );
    glLightfv( GL_LIGHT1, GL_POSITION, Globals::light1_pos );
}

//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
    
    switch (Globals::interactionMode){
            
        case (MODE_ENTERWORD):
        {
            
            switch ( int(key) ){
                case (8):
                case (127):
                    Globals::controlPanel->erase(Globals::interactionMode); break;
                case (32):
                case (13):
                {
                    
                    // turn notes off
                    for (int i = 0; i < NNOTES; i++){
                        Globals::synth.at(i)->adsrGain.keyOff();
                        Globals::synth.at(i)->adsrFreq.keyOff();
                        Globals::synth.at(i)->adsrRad.keyOff();
                        Globals::synth.at(i)->fm.keyOff();
                    }
                    
                    Globals::timeSinceLastInteraction = 0;
                    Globals::wordList->resetRandFreqs();
                    
                    // clear word list
                    string words = Globals::controlPanel->getWords();
                    for (int k = 0; k < words.length(); k++) Globals::controlPanel->erase(MODE_ENTERWORDS);
                    
                    // modify synthesizer quality & confdience of existing word
                    if (Globals::focalWord != 0x0) Globals::focalWord->updateConfidence();
                    // modify tempo & frequencies of existing word
                    if (Globals::focalWord) {
                        Globals::focalWord->setTempo(Globals::eighthNoteTempo);
                        for (int i = 0 ; i < NPASTFREQS; i++) {
                            Globals::focalWord->setFreqs(i, Globals::pastFreqs[i]);
                        }
                    }
                    
                    // examine new word
                    string newWordText = Globals::controlPanel->getWord();
                    int isListed = Globals::wordList->isListed(newWordText);
                    switch (isListed){
                        case (0):
                        {
                            wsWord * newWord = new wsWord();
                            // set string
                            newWord->setString(newWordText, (int) newWordText.length());
                            // reset confidence, number of tests
                            newWord->resetConfidence(0, 0);
                            // randomize color, for now
                            newWord->setColor(Vector3D((float)rand()/RAND_MAX,(float)rand()/RAND_MAX,(float)rand()/RAND_MAX));
                            // set random note
                            // add word to new word list
                            Globals::wordList->addNew(newWord);
                            // change focal point of word list
                            Globals::focalWord = Globals::wordList->newWordsMap.find(newWordText)->second;
                            
                            // add word to simulation
                            Globals::sim->root().addChild(Globals::focalWord->yText);
                            break;
                        }
                            
                        case (1):
                        {
                            // change focal point of word list
                            Globals::focalWord = Globals::wordList->existingWordsMap.find(newWordText)->second;
                            break;
                        }
                        case (2):
                        {
                            // change focal point of word list
                            Globals::focalWord = Globals::wordList->newWordsMap.find(newWordText)->second;
                            break;
                        }
                        case(3):
                        {

                            // change focal point of word list
                            Globals::focalWord = Globals::wordList->editedWordsMap.find(newWordText)->second;
                            Globals::wordList->recomputePos(Globals::focalWord);
                            break;
                        }
                    }
                    
                    // update tempo & frequencies
                    Globals::eighthNoteTempo = Globals::focalWord->getTempo();
                    for (int i = 0; i < NPASTFREQS; i++) {
                        Globals::pastFreqs[i] = Globals::focalWord->getFreqs(i);
                        if (Globals::pastFreqs[i] == 0) Globals::pastFreqs[i] = 440;
                    }
                    // randomize note frequencies
                    Globals::wordList->resetRandFreqs();
                    
                    
                    // initialize confidence of new focal word
                    Globals::focalWord->initializeConfidence();
                    // compute position
                    Globals::wordList->recomputePos(Globals::focalWord);
                    Globals::controlPanel->setCirclePos(Globals::focalWord->getConfidence());
                    for (int i = 0; i < NNOTES; i++) Globals::focalWord->setSynth(Globals::synth.at(i));
                    // reset random note
                    
                    Vector3D pos;
                    float val = float(rand())/RAND_MAX-.5;
                    val = val < 0 ? floor(val) : ceil(val);
                    Globals::viewEyeX.value = Globals::viewEyeX.value + val*(2*PI)*round(5*float(rand())/RAND_MAX+1);
                    Globals::focalWord->getPosFinal(& pos);
                    Globals::viewX.update(pos[0],VIEW_SLEWRATE);
                    Globals::viewY.update(pos[1],VIEW_SLEWRATE);
                    Globals::viewZ.update(pos[2], VIEW_SLEWRATE);
                    Globals::viewEyeX.y = 0;
                    Globals::viewEyeY.y = 0;
                    
                    break;
                }
                
                default: Globals::controlPanel->write(Globals::interactionMode, key); break;
                    
            }
            break;
        }
            
        case (MODE_ENTERWORDS):
        {
            
            switch ( int(key) ){
                case (8):
                case (127):{
                    Globals::controlPanel->erase(Globals::interactionMode);
                    break;
                }
                case (13):{
                    string wordList = Globals::controlPanel->getWords();
                    Globals::wordList->addList(wordList, Globals::focalWord);
                    cerr << wordList << endl;
                    for (int i = 0; i < wordList.length(); i++) Globals::controlPanel->erase(MODE_ENTERWORDS);
                    break;
                }
                    
                default: Globals::controlPanel->write(Globals::interactionMode, key); break;
            }
            break;
        }
        case (MODE_MOVECIRCLE):
        case (MODE_NORMAL):
        {
    
            switch( key )
            {
                case 'q': {
                    Globals::audioIO->stopStream();
                    Globals::audioIO->closeStream();
                    if (Globals::focalWord != 0x0) Globals::focalWord->updateConfidence();
                    Globals::wordList->save(Globals::wordFile);
                    Globals::wordList->savePCASettings(Globals::pcaFile);
                    exit(0); break;
                }
                case 'h': help(); break;
                case 's':
                {
                    if( !Globals::fullscreen )
                    {
                        Globals::lastWindowWidth = Globals::windowWidth;
                        Globals::lastWindowHeight = Globals::windowHeight;
                        glutFullScreen();
                    }
                    else
                        glutReshapeWindow( Globals::lastWindowWidth, Globals::lastWindowHeight );
            
                    Globals::fullscreen = !Globals::fullscreen;
                    break;
                }
                case 'r':
                {
                    Globals::wordList->recomputePos();
                    Globals::timeSinceLastInteraction = 0;
                    
                    Vector3D pos;
                    Globals::focalWord->getPosFinal(& pos);
                    Globals::viewX.update(-10*float(rand())/RAND_MAX+5,VIEW_SLEWRATE_SLOW);
                    Globals::viewY.update(-10*float(rand())/RAND_MAX+5,VIEW_SLEWRATE_SLOW);
                    Globals::viewZ.update(5,VIEW_SLEWRATE_SLOW);
                    Globals::viewEyeX.y = 0;
                    Globals::viewEyeY.y = 0;
                    
                    Globals::timeSinceLastR = 0;
                    Globals::hasExploded = 0;
                    
                    Globals::ringingWord = Globals::focalWord;
                    
                    break;
                }
                case ']':
                    if (Globals::viewEyeY.y > -G_PI/2.1) Globals::viewEyeY.y -= .1f;
                    break;
                case '[':
                    if (Globals::viewEyeY.y < G_PI/2.1) Globals::viewEyeY.y += .1f;
                    break;
                case 'p':
                    Globals::viewEyeX.y += .1f;
                    break;
                case '\\':
                    Globals::viewEyeX.y -= .1f;
                    break;
                case 'l':
                    Globals::focalWord->setTempo(Globals::focalWord->getTempo() + 4);
                    break;
                case 'k':
                    Globals::focalWord->setTempo(Globals::focalWord->getTempo() - 4);
                    if (Globals::focalWord->getTempo() <= 0) Globals::focalWord->setTempo(4);
                    break;
                case 'm':
                    Globals::noteDurationHoldTypical += 1;
                    break;
                case 'n':
                    Globals::noteDurationHoldTypical -= 1;
                    if (Globals::noteDurationHoldTypical < 2) Globals::noteDurationHoldTypical = 2;
                    break;
            }
        }
    }

    // do a reshape since viewEyeY might have changed
    reshapeFunc( Globals::windowWidth, Globals::windowHeight );
    // post redisplay
    glutPostRedisplay( );
}


//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
    
    switch( button )
    {
            
        // left mouse button
        case 0:
            
            float x_ = float(2*x)/float(Globals::windowWidth)-1;
            float y_ = -float(2*y)/float(Globals::windowHeight)+1;
            int state = Globals::controlPanel->isInObject(x_, y_);
            
            switch (state){
                case (MODE_NORMAL): Globals::interactionMode = MODE_NORMAL; break;
                case(MODE_MOVECIRCLE): Globals::interactionMode = MODE_MOVECIRCLE; break;
                case(MODE_ENTERWORD): Globals::interactionMode = MODE_ENTERWORD; break;
                case(MODE_ENTERWORDS): Globals::interactionMode = MODE_ENTERWORDS; break;
            }

    }
    glutPostRedisplay( );
}

// -----------------
// Active Mouse Func
// -----------------
void activeMouseFunc( int x, int y){
    
    if (Globals::interactionMode == MODE_MOVECIRCLE){
        float x_ = float(2*x)/float(Globals::windowWidth)-1;
        float y_ = -float(2*y)/float(Globals::windowHeight)+1;
        Globals::controlPanel->setCirclePos(x_,y_);
        if (Globals::focalWord != 0x0) Globals::focalWord->setConfidence( Globals::controlPanel->getCirclePos() );
    }
    
}

// ------------------
// Passive Mouse Func
// ------------------
void passiveMouseFunc (int x, int y){
    
    if (y > .8 * Globals::windowHeight) Globals::controlPanel->slideUp();
    else {
        Globals::controlPanel->slideDown();
        Globals::interactionMode = MODE_NORMAL;
    }
}

//-----------------------------------------------------------------------------
// Name: specialFunc( )
// Desc: handles arrow stuff
//-----------------------------------------------------------------------------
void specialFunc( int key, int x, int y )
{
    // check
    bool handled = false;
    
    // if not handled
    if( !handled )
    {
        switch( key )
        {
            case GLUT_KEY_LEFT:
                break;
            case GLUT_KEY_RIGHT:
                break;
            case GLUT_KEY_UP:
                break;
            case GLUT_KEY_DOWN:
                break;
        }
    }
}




//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
    // render the scene
    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
    
    // update times
    XGfx::getCurrentTime( TRUE );
    Globals::timeSinceLastR = Globals::timeSinceLastR + XGfx::delta();
    Globals::timeSinceLastInteraction = Globals::timeSinceLastInteraction + XGfx::delta();
    Globals::timeSinceLastRing = Globals::timeSinceLastRing + XGfx::delta();
    
    
    if (Globals::focalWord) {
    
        // check for explosion
        float dist = Globals::wordList->getMaxDist();
        if ( ((Globals::timeSinceLastR > 20) || dist < 1) && (Globals::hasExploded==0)){
            
            Globals::viewEyeX.y = 0;
            
            Globals::hasExploded = 1;
            Globals::wordList->setPosGoals();
            
            Vector3D pos;
            Globals::focalWord->getPosFinal(& pos);
            Globals::viewX.update(pos[0],VIEW_SLEWRATE);
            Globals::viewY.update(pos[1],VIEW_SLEWRATE);
            Globals::viewZ.update(pos[2],VIEW_SLEWRATE);
            
        }
        else if (Globals::hasExploded==0) { Globals::timeSinceLastInteraction = 0; }
        
        float meanDist = Globals::wordList->getMeanDistFromFinal();
        
        Globals::eighthNoteTempo = Globals::focalWord->getTempo() + meanDist;
        
        if (Globals::hasExploded==1 && meanDist <= 10){
            
            Globals::hasExploded = 2;
            
            // turn notes off
            for (int i = 0; i < NNOTES; i++){
                Globals::synth.at(i)->adsrGain.keyOff();
                Globals::synth.at(i)->adsrFreq.keyOff();
                Globals::synth.at(i)->adsrRad.keyOff();
                Globals::synth.at(i)->fm.keyOff();
            }
        }
        
        if ( (Globals::hasExploded == 1) && (Globals::inPlay == 0) && (Globals::timeSinceLastRing > 2/Globals::eighthNoteTempo ) ){
            
            // turn off sounds after a quarter beat
            if (Globals::timeSinceLastRing > 2/Globals::eighthNoteTempo){
                
                // turn notes off
                for (int i = 0; i < NNOTES; i++){
                    Globals::synth.at(i)->adsrGain.keyOff();
                    Globals::synth.at(i)->adsrFreq.keyOff();
                    Globals::synth.at(i)->adsrRad.keyOff();
                    Globals::synth.at(i)->fm.keyOff();
                }
                
                // set duration
                Globals::noteDuration = 2/Globals::eighthNoteTempo;
                
                // play!
                Globals::synth.at(Globals::synthInd)->adsrGain.keyOn();
                Globals::synth.at(Globals::synthInd)->adsrFreq.keyOn();
                Globals::synth.at(Globals::synthInd)->adsrRad.keyOn();
                Globals::synth.at(Globals::synthInd)->fm.noteOn(Globals::pastFreqs[0] + pow(meanDist,1.5), .01);
                
                // change status
                Globals::isRinging = 1;
                Globals::timeSinceLastRing = 0;
                
            }
            
            Globals::timeSinceLastInteraction = 0;
            
        }
        
        // randomly play sounds [if the time since your last interaction is greater than 2]
        if ( (((Globals::timeSinceLastInteraction > 5) && (Globals::hasExploded == 2)) || (Globals::hasExploded == 0) ) && (Globals::inPlay == 0) && (Globals::timeSinceLastRing > 2/Globals::eighthNoteTempo )){
            
            
            // turn off sounds after a quarter beat
            if (Globals::timeSinceLastRing > Globals::noteDurationHold){
                // turn notes off
                for (int i = 0; i < NNOTES; i++){
                    Globals::synth.at(i)->adsrGain.keyOff();
                    Globals::synth.at(i)->adsrFreq.keyOff();
                    Globals::synth.at(i)->adsrRad.keyOff();
                    Globals::synth.at(i)->fm.keyOff();
                }
                
            }
            
            // select a note, possibly
            // ---------------------------------------------------------------------------------------------------------------------------------
            if (Globals::timeSinceLastRing > Globals::noteDuration){
                
                Vector3D pos1, pos2;
                float pNoteOn, pNoteOnFinal;
                Globals::focalWord->getPosFinal(& pos1);
                
                pNoteOnFinal = 0;
                
                for (auto& a: Globals::wordList->existingWordsMap){
                    
                    a.second->getPosFinal(&pos2);
                    dist = pow(pos1.x-pos2.x,2) + pow(pos1.y-pos2.y,2) + pow(pos1.z-pos2.z,2);
                    if (dist > .0005 && dist < 500){
                        
                        pNoteOn = (float(rand())/RAND_MAX)*(1/(1+exp(-10*(Globals::timeSinceLastInteraction/10 + .5))))*(1-dist/500);
                        
                        if (pNoteOn > pNoteOnFinal){
                            Globals::ringingWord = a.second;
                            pNoteOnFinal = pNoteOn;
                        }
                        
                    }
                }
                for (auto& a: Globals::wordList->editedWordsMap){
                    
                    a.second->getPosFinal(&pos2);
                    dist = pow(pos1.x-pos2.x,2) + pow(pos1.y-pos2.y,2) + pow(pos1.z-pos2.z,2);
                    if (dist > .0005 && dist < 500){
                        
                        pNoteOn = (float(rand())/RAND_MAX)*(1/(1+exp(-10*(Globals::timeSinceLastInteraction/10 + .5))))*(1-dist/500);
                        
                        if (pNoteOn > pNoteOnFinal){
                            Globals::ringingWord = a.second;
                            pNoteOnFinal = pNoteOn;
                        }
                        
                    }
                }
                for (auto& a: Globals::wordList->newWordsMap){
                    
                    a.second->getPosFinal(&pos2);
                    dist = pow(pos1.x-pos2.x,2) + pow(pos1.y-pos2.y,2) + pow(pos1.z-pos2.z,2);
                    if (dist > .0005 && dist < 500){
                        
                        pNoteOn = (float(rand())/RAND_MAX)*(1/(1+exp(-10*(Globals::timeSinceLastInteraction/10 + .5))))*(1-dist/500);
                        
                        if (pNoteOn > pNoteOnFinal){
                            Globals::ringingWord = a.second;
                            pNoteOnFinal = pNoteOn;
                        }
                        
                    }
                }
                // ---------------------------------------------------------------------------------------------------------------------------------
                
                // make word jump
                if (Globals::ringingWord && (Globals::hasExploded == 2)){
                    Globals::ringingWord->getPos(&pos2);
                    Globals::ringingWord->setPosNow(Vector3D(pos2.x,pos2.y+.5,pos2.z));
                }
                
                // set random duration
                Globals::noteDuration = 8*pow(2,floor( 4*float(rand())/RAND_MAX ))/Globals::eighthNoteTempo;
                Globals::noteDurationHold = ceil(Globals::noteDurationHoldTypical*.75) + (float(rand())/RAND_MAX)*(Globals::noteDurationHoldTypical*.5)/Globals::eighthNoteTempo;
                
                cerr << "note duration: " << Globals::noteDurationHold << endl;
                cerr << "tempo: " << Globals::eighthNoteTempo << endl;
                Globals::noteDurationHold = (Globals::noteDurationHold < 2/Globals::eighthNoteTempo)? 2/Globals::eighthNoteTempo:Globals::noteDurationHold;
                
                // turn notes off
                for (int i = 0; i < NNOTES; i++){
                    Globals::synth.at(i)->adsrGain.keyOff();
                    Globals::synth.at(i)->adsrFreq.keyOff();
                    Globals::synth.at(i)->adsrRad.keyOff();
                    Globals::synth.at(i)->fm.keyOff();
                }
                
                // play!
                Globals::synth.at(Globals::synthInd)->adsrGain.keyOn();
                Globals::synth.at(Globals::synthInd)->adsrFreq.keyOn();
                Globals::synth.at(Globals::synthInd)->adsrRad.keyOn();
                if (Globals::ringingWord) Globals::synth.at(Globals::synthInd)->fm.noteOn(Globals::ringingWord->getRandFreq() + pow(meanDist,1.5), .01);
                
                // change status
                Globals::isRinging = 1;
                Globals::timeSinceLastRing = 0;
                
            }
            
        }
        
    }
    
    // ---------------------------------------------------------------------------------------------------------------------------------
    
    // update positions
    Globals::bgColor.interp( XGfx::delta() );
    Globals::blendAlpha.interp( XGfx::delta() );
    
    Globals::wordList->interp( XGfx::delta() );
    Globals::wordList->updatePositions();
    
    if (Globals::inPlay==1) Globals::focalWord->updatePos(*Globals::focalWordTrans);
    else if (Globals::ringingWord && (Globals::inPlay==0) ) Globals::ringingWord->updatePos(*Globals::focalWordTrans,.2);
    
    // clear or blend
    if( Globals::blendScreen && Globals::blendAlpha.value > .0001 )
    {
        // clear the depth buffer
        glClear( GL_DEPTH_BUFFER_BIT );
        // blend screen
        blendPane();
    }
    else
    {
        // set the GL clear color - use when the color buffer is cleared
        glClearColor( Globals::bgColor.actual().x, Globals::bgColor.actual().y, Globals::bgColor.actual().z, 1.0f );
        // clear the color and depth buffers
        glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    }
    
    // enable depth test
    glEnable( GL_DEPTH_TEST );
    
    renderBackground();
    
    // save state
    glPushMatrix();
    
    // slew
    Globals::viewEyeY.interp( XGfx::delta());
    Globals::viewEyeX.interp( XGfx::delta());
    Globals::viewRadius.interp(XGfx::delta());
    Globals::viewX.interp( XGfx::delta());
    Globals::viewY.interp( XGfx::delta());
    Globals::viewZ.interp( XGfx::delta());
    
    look(W_PERS);

    // cascade simulation
    Globals::sim->systemCascade();
    
    // pop state
    glPopMatrix();
    
    //    // draw any HUD here
    //    Globals::hud->project();
    //    Globals::hud->updateAll( Globals::sim->delta() );
    //    Globals::hud->drawAll();
    
    // flush gl commands
    glFlush();
    // swap the buffers
    glutSwapBuffers();
}

//-----------------------------------------------------------------------------
// name: blendPane()
// desc: blends a pane into the current scene
//-----------------------------------------------------------------------------
void blendPane()
{
    // enable blending
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    // disable lighting
    glDisable( GL_LIGHTING );
    // disable depth test
    glDisable( GL_DEPTH_TEST );
    // blend in a polygon
    glColor4f( Globals::bgColor.actual().x, Globals::bgColor.actual().y, Globals::bgColor.actual().z, Globals::blendAlpha.value );
    
    GLfloat h = 10;
    GLfloat d = -1;
    
    // draw the polyg
    glBegin( GL_QUADS );
    glVertex3f( -h, -h, d );
    glVertex3f( h, -h, d );
    glVertex3f( h, h, d );
    glVertex3f( -h, h, d );
    glEnd();
    
    // enable lighting
    glEnable( GL_LIGHTING );
    // enable depth test
    glEnable( GL_DEPTH_TEST );
    // disable blending
    glDisable( GL_BLEND );
}

//-----------------
// renderBackground
//-----------------
void renderBackground()
{
    // enable state
    glEnableClientState( GL_VERTEX_ARRAY );
    glEnableClientState( GL_NORMAL_ARRAY );
    
    // push
    glPushMatrix();
    
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(100, 100, 0, 100);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glColor3f(1, 1, 1);
    
    glBegin(GL_QUADS);
    glColor4d(0,0,.1,1);
    glVertex2f(-1,-1);
    glColor4d(.2,0,.1,1);
    glVertex2f(1,-1);
    glColor4d(0,.2,.2,1);
    glVertex2f(1,1);
    glColor4d(0,.2,0,1);
    glVertex2f(-1,1);
    glEnd();
    
    // pop
    glPopMatrix();
    
    // disable
    glDisableClientState( GL_VERTEX_ARRAY );
    glDisableClientState( GL_NORMAL_ARRAY );
}


// ---------------------
// save past frequencies
// ---------------------
void savePastFreqs(string fileName){
    
    ofstream myOut;
    myOut.open(Globals::tempFile);
    string myLine, myWord;
    
    for (int i = 0; i < NPASTFREQS; i++){
        cerr << Globals::pastFreqs[i] << endl;
        myOut << Globals::pastFreqs[i] << ",";
    }
    myOut << endl;
    
    myOut.close();
    
    // move temporary file to actual file
    rename( Globals::tempFile.c_str(), fileName.c_str() );
    
}

// ---------------------
// load past frequencies
// ---------------------
void loadPastFreqs(string fileName){
    
    ifstream myIn;
    string myLine, myWord;
    int ind = 0;
    
    myIn.open(fileName.c_str());
    
    getline(myIn, myLine);
    
    myWord.clear();
    
    for (int i = 0; i < myLine.length(); i++){
        
        if (myLine[i]==','){
            Globals::pastFreqs[ind] = atof(myWord.c_str());
            myWord.clear();
        }
        else{
            myWord.append(myLine.substr(i,1));
        }
        
    }
    
    for (int i = ind; i < NPASTFREQS; i++){
        Globals::pastFreqs[i] = 440;
    }
    
    myIn.close();
    
}