#include <math.h>
#include <iostream>
#include <cstdlib>
using namespace std;

#include "RtAudio.h"
#include "RtMidi.h"

#include "globals.h"
#include "gfx.h"

// sample datatype
#define SAMPLE double
// RtAudio format
#define FORMAT RTAUDIO_FLOAT64
// pi!
#define PI 3.14159265358979

// ----------------------
// name: audioIOCallback()
// desc: audio callback
// ----------------------
int audioIOCallback( void * outputBuffer, void * inputBuffer, unsigned int numFrames, double streamTime, RtAudioStreamStatus status, void * data)
{
    
    StkFrames fm(numFrames, 1);
    StkFrames adsr(numFrames, 1);
    StkFrames filtered(numFrames, 1);
    StkFrames freq(numFrames, 1);
    StkFrames rad(numFrames, 1);
    SAMPLE * buffy = (SAMPLE *)outputBuffer;
    
    // synthesize
    Globals::synth.at(0)->fm.tick(fm);
    // gain
    Globals::synth.at(0)->adsrGain.tick(adsr);
    for (int i = 0; i < numFrames; i++) filtered[i] = .5*adsr[i]*fm[i];
    // resonance
    Globals::synth.at(0)->adsrFreq.tick(freq);
    Globals::synth.at(0)->adsrRad.tick(rad);
    Globals::synth.at(0)->filter.setResonance(freq[0], rad[0]);
    Globals::synth.at(0)->filter.tick(filtered);
    // echo
    Globals::synth.at(0)->echo.tick(filtered);
    // output
    for (int i = 0; i < numFrames; i++) for (int j = 0; j < NCHANNELS; j++) buffy[i*NCHANNELS+j] = (SAMPLE) filtered[i];
    
    // vibrate focal word
    Globals::focalWordTrans->x = .5*buffy[0];
        
    return 0;
    
}

// --------------
// midiInCallback
// --------------
void midiInCallback( double deltatime, std::vector< unsigned char > *message, void *userData )
{
 
    Globals::timeSinceLastInteraction = 0;
    
    // randomize note frequencies
    Globals::wordList->resetRandFreqs();
    
    // play
    if ((int)message->at(0)==144 && (int)message->at(2)!=0)
    {

        Globals::inPlay = 1;
        
        // turn notes off
        for (int i = 0; i < NNOTES; i++){
            Globals::synth.at(i)->adsrGain.keyOff();
            Globals::synth.at(i)->adsrFreq.keyOff();
            Globals::synth.at(i)->adsrRad.keyOff();
            Globals::synth.at(i)->fm.keyOff();
        }
        
        // compute frequency
        float freq = 440*pow(2, ((double)message->at(1)-57)/12);

        // ADSR
        Globals::synth.at(Globals::synthInd)->adsrGain.keyOn();
        Globals::synth.at(Globals::synthInd)->adsrFreq.keyOn();
        Globals::synth.at(Globals::synthInd)->adsrRad.keyOn();
        
        // play!
        Globals::synth.at(Globals::synthInd)->fm.noteOn(freq, .1);
        
        // update list of previous notes
        Globals::pastFreqs[Globals::pastFreqsInd] = freq;
        Globals::pastFreqsInd = (Globals::pastFreqsInd + 1)%NPASTFREQS;
        
        // make focal word jump
        Vector3D *pos = new Vector3D();
        Globals::focalWord->getPos(pos);
        Globals::focalWord->setPosNow(Vector3D(pos->x,pos->y+float((int)message->at(2))/512,pos->z));
        

    }
    // stop
    else if ((int) message->at(0)==128 || ((int) message->at(0)==144 && (int) message->at(2)==0))
    {
        for (int i = 0; i < NNOTES; i++){
            Globals::synth.at(i)->adsrGain.keyOff();
            Globals::synth.at(i)->adsrFreq.keyOff();
            Globals::synth.at(i)->adsrRad.keyOff();
            Globals::synth.at(i)->fm.keyOff();
        }
        
        // make focal word dip
        Vector3D *pos = new Vector3D();
        Globals::focalWord->getPos(pos);
        Globals::focalWord->setPosNow(Vector3D(pos->x,pos->y-float((int)message->at(2))/512,pos->z));
        
        Globals::inPlay = 0;
        Globals::timeSinceLastInteraction = 0;

        
    }

}

// ----------------------
// name: main()
// desc: entry pointr
// ----------------------
int main( int argc, const char ** argv )
{
    
    // invoke graphics setup and loop;
    if( !gfx_init( argc, argv ) )
    {
        // error message
        cerr << "cannot initialize graphics/data system..." << endl;
        return -1;
    }
    
    // ---------------
    // INITIALIZE MIDI
    // ---------------
    int iMidiIn, nPorts;
    string portName;
    ////
    // create new midi in object
    try {
        Globals::midiIn = new RtMidiIn();
    }
    catch (RtError &error){
        error.printMessage();
    }

    // check inputs
    nPorts = Globals::midiIn->getPortCount();
    // exit if no ports are available
    if (nPorts==0){
         cout << "No ports available! : (" << endl;
    }
    else{
        // list input ports
        cout << "\n";
        for (unsigned int i=0; i<nPorts; i++){
            try{
                portName = Globals::midiIn->getPortName(i);
            }
            catch( RtError &error){
                error.printMessage();
            }
            cout << i+1 << ": " << portName << endl;
        }
        
        int catcher = 0;
        while (catcher == 0){
            
            try {
                
                catcher = 1;
                
                // get user to select input
                cout << "Please select 1 of " << nPorts << " MIDI input source[s].\n";
                cin >> iMidiIn;
        
                // open port
                Globals::midiIn->openPort(iMidiIn-1);
            }
            catch (RtError &error){
                catcher = 0;
            }
            
        }
        // set callback
        Globals::midiIn->setCallback( &midiInCallback );
        // don't ginore syex, timing, or active sensing messages
        Globals::midiIn->ignoreTypes(false,false,false);
    }

    
    // ----------------------
    // Initialize Synthesizer
    // ----------------------
    
    // --------
    // AUDIO IO
    // --------
    // instantiate RtAudio object
    Globals::audioIO = new RtAudio();
    Globals::iStreamParams = new RtAudio::StreamParameters;
    Globals::oStreamParams = new RtAudio::StreamParameters;
    
    // check for audio devices
    if ( Globals::audioIO->getDeviceCount() < 1)
    {
        // nopes
        cout << "no audio devices found! :(" << endl;
        exit(EXIT_FAILURE);
    }
    
    // let RtAudio print messages to stderr
    Globals::audioIO->showWarnings(true);
    
    // set input and output parameters
    Globals::iStreamParams->deviceId = Globals::audioIO->getDefaultInputDevice();
    Globals::iStreamParams->nChannels = NCHANNELS;
    Globals::iStreamParams->firstChannel = 0;
    Globals::oStreamParams->deviceId = Globals::audioIO->getDefaultOutputDevice();
    Globals::oStreamParams->nChannels = NCHANNELS;
    Globals::oStreamParams->firstChannel = 0;
    
    // create stream options
    RtAudio::StreamOptions options;
    
    // open stream
    try {
        Globals::audioIO->openStream( Globals::oStreamParams, Globals::iStreamParams, FORMAT, G_SRATE, &Globals::bufferFrames, & audioIOCallback, (void*)& Globals::bufferBytes, &options);
    }
    catch (RtError & e)
    {
        // error!
        cout << e.getMessage() << endl;
        exit(EXIT_FAILURE);
    }
    
    // compute number of bytes/buffer
    Globals::bufferBytes = Globals::bufferFrames * NCHANNELS * sizeof(SAMPLE);
    
    // start stream
    try {
        Globals::audioIO->startStream();
    }
    catch (RtError& e)
    {
        // error!
        cout << e.getMessage() << endl;
    }
    
    // -------------
    // GRAPHICS LOOP
    // -------------
    gfx_loop();
    
    
}